@@ -22,7 +22,10 @@
 const char *malloc_conf = NULL;
 
 /* Without jemalloc, isc_mem_get_align() is equal to isc_mem_get() */
-#define MALLOCX_ALIGN(a) (a & 0) /* Clear the flag */
+#define MALLOCX_ALIGN(a)    (a & 0) /* Clear the flag */
+#define MALLOCX_ZERO	    ((int)0x40)
+#define MALLOCX_TCACHE_NONE (0)
+#define MALLOCX_ARENA(a)    (0)
 
 #if defined(HAVE_MALLOC_SIZE) || defined(HAVE_MALLOC_USABLE_SIZE)
 
@@ -64,7 +67,15 @@ sallocx(void *ptr, int flags) {
 
 #elif HAVE_MALLOC_USABLE_SIZE
 
+#ifdef __DragonFly__
+/*
+ * On DragonFly BSD 'man 3 malloc' advises us to include the following
+ * header to have access to malloc_usable_size().
+ */
+#include <malloc_np.h>
+#else
 #include <malloc.h>
+#endif
 
 static inline size_t
 sallocx(void *ptr, int flags) {
